@@ -44,11 +44,13 @@ try:
     from common import TS3Error
     from escape import TS3Escape
     from response import TS3Response, TS3QueryResponse, TS3Event
+    import _lib
 except ImportError:
     from .commands import TS3Commands
     from .common import TS3Error
     from .escape import TS3Escape
     from .response import TS3Response, TS3QueryResponse, TS3Event
+    from . import _lib
 
 
 # Backward compatibility
@@ -149,11 +151,22 @@ class TS3BaseConnection(object):
         # Avoid sending data to the same time.
         self._send_lock = threading.Lock()
 
+        # Needed to store a link to the keep alive thread
+        # and information about the keep alive intervall.
+        self._keepalive_timer = None
+        self._keepalive_interval = None
+
         # To stop the receive progress, if we are waiting for events from
         # another thread.
         self._stop_event = threading.Event()
         self._waiting_for_stop = False
         self._is_listening = False
+
+        # Dispatches all received events in another thread to avoid
+        # dead locks.
+        # The dispatcher is started, as soon as the first event
+        # is received and stopped, when the connection is closed.
+        self.__event_dispatcher = _lib.SignalDispatcher()
         
         if host is not None:
             self.open(host, port)
@@ -258,8 +271,9 @@ class TS3BaseConnection(object):
                 self._telnet_conn.write(b"quit\n\r")
             finally:
                 self.stop_recv()
+                self.cancel_keepalive()
                 self._telnet_conn.close()
-                self._telnet_conn = None
+                self._telnet_conn = None                
                 _logger.debug("Disconnected client.")
         return None
 
@@ -283,6 +297,72 @@ class TS3BaseConnection(object):
         self.close()
         return None
 
+    # Keep alive
+    # -------------------------
+    
+    def __send_keepalive_beacon(self):
+        """
+        Sends the keep alive beacon ``\n\r`` to the server and restarts the
+        keep alive timer :attr:`_keepalive_timer`.
+
+        .. seealso::
+
+            * :meth:`cancel_keepalive`
+            * :meth:`keepalive`
+        """        
+        # Send the beacon.
+        with self._send_lock:
+            self._telnet_conn.write(b"\n\r")
+
+        # Restart the timer.
+        if self._keepalive_timer is not None:
+            self._keepalive_timer.cancel()
+            self._keepalive_timer = None
+
+        if self._keepalive_interval is not None:            
+            self._keepalive_timer = threading.Timer(
+                interval = self._keepalive_interval,
+                function = self.__send_keepalive_beacon
+                )
+            self._keepalive_timer.start()
+        return None
+
+    def cancel_keepalive(self):
+        """
+        Cancels the *keepalive* beacon started using :meth:`keepalive`.
+
+        .. seealso::
+
+            * :meth:`keepalive`
+        """
+        if self._keepalive_timer is not None:
+            self._keepalive_timer.cancel()
+            self._keepalive_timer = None
+
+        self._keepalive_interval = None
+        return None            
+
+    def keepalive(self, interval=540):
+        """
+        Starts or restarts a timer which sends each *interval* seconds a beacon
+        to the ts3 server to prevent closing the connection due to the max idle
+        time.
+
+        :arg interval:
+            Time between a beacon is sent to the server. This should be less than
+            the maximum allowed idle time. Usually, this value is set to 600s on
+            a TS3 server. 
+
+        .. seealso::
+
+            * :meth:`cancel_keepalive`
+        """
+        # Sending a beacon now, will create a new timer
+        # or restart the timer with the new intervall.
+        self._keepalive_interval = interval
+        self.__send_keepalive_beacon()
+        return None
+
     # Receiving
     # -------------------------
     
@@ -370,6 +450,7 @@ class TS3BaseConnection(object):
         if self._is_listening:
             self._stop_event.clear()
             self._waiting_for_stop = True
+            self.__event_dispatcher.stop()
             self._stop_event.wait()
         return None
 
@@ -387,11 +468,15 @@ class TS3BaseConnection(object):
         Blocks untill all unfetched responses have been received or
         forever, if *recv_forever* is true.
 
+        .. deprecated:: 0.6.0
+            The *recv_forever* argument will be removed in future versions.
+            Use :meth:`recv_in_thread` instead.
+            
         :arg recv_forever:
             If true, this method waits forever for a response and you have to
             call :meth:`stop_recv` from another thread, to stop it.
         :type recv_forever:
-            bool
+            bool           
 
         :arg poll_intervall:
             Seconds between checks for the stop request.
@@ -427,7 +512,13 @@ class TS3BaseConnection(object):
                 # We received an event.
                 if data.startswith(b"notify"):
                     event = TS3Event([data])
-                    self.on_event.send(self, event=event)
+
+                    # We start the event dispatcher as late as possible to
+                    # avoid unnecessairy threads.
+                    self.__event_dispatcher.start()
+                    self.__event_dispatcher.send(
+                        self.on_event, self, event=event
+                        )
                     
                 # We received the end of a query response.
                 elif data.startswith(b"error"):
